import astropy.units as u
import numpy as np
import healpy as hp
import copy

import logging
logger = logging.getLogger(__name__)

from histpy import Histogram, Axes, Axis, HealpixAxis

from cosipy.response.functions import get_integrated_spectral_model

from .model_base import ModelBase

class AllSkyImageModel(ModelBase):
    """
    Photon flux maps in given energy bands. 2-dimensional histogram.

    Attributes
    ----------
    nside : int
        Healpix NSIDE parameter.
    energy_edges : :py:class:`np.array`
        Bin edges for energies. We recommend to use a Quantity array with the unit of keV.
    scheme : str, default 'ring'
        Healpix scheme. Either 'ring', 'nested'.
    coordsys : str or :py:class:`astropy.coordinates.BaseRepresentation`, default is 'galactic'
        Instrinsic coordinates of the map. The default is 'galactic'.
    label_image : str, default 'lb'
        The label name of the healpix axis.
    label_energy : str, default 'Ei'
        The label name of the energy axis. The default is 'Ei'.
    """

    def __init__(self,
                 nside,
                 energy_edges,
                 scheme = 'ring',
                 coordsys = 'galactic',
                 label_image = 'lb',
                 label_energy = 'Ei',
                 unit = '1/(s*cm*cm*sr)'
                 ):

        if energy_edges.unit != u.keV:

            logger.warning(f"The unit of the given energy_edges is {energy_edges.unit}. It will be converted to keV.")
            energy_edges = energy_edges.to('keV')

        image_axis = HealpixAxis(nside = nside,
                                 scheme = scheme,
                                 coordsys = coordsys,
                                 label = label_image)

        energy_axis = Axis(edges = energy_edges, label = label_energy, scale = "log")

        axes = Axes([image_axis, energy_axis])

        super().__init__(axes, sparse = False, unit = unit)

    @classmethod
    def open(cls, filename, name = 'hist'):
        """
        Open a file

        Parameters
        ----------
        filename: str
        
        Returns
        -------
        py:class:`AllSkyImageModel`
        """

        hist = Histogram.open(filename, name)

        allskyimage = AllSkyImageModel(nside = hist.axes[0].nside, 
                                    energy_edges = hist.axes[1].edges,
                                    scheme = hist.axes[0].scheme, 
                                    coordsys = hist.axes[0].coordsys.name, 
                                    label_image = hist.axes[0].label, 
                                    label_energy = hist.axes[1].label,
                                    unit = hist.unit)

        allskyimage[:] = hist.contents

        del hist
        return allskyimage

    @classmethod
    def instantiate_from_parameters(cls, parameter):
        """
        Generate an instance of this class

        Parameters
        ----------
        parameter : py:class:`yayc.Configurator`
        
        Returns
        -------
        py:class:`AllSkyImageModel`

        Notes
        -----
        The parameters should be given like this:

        nside: 8
        energy_edges:
            value: [100.,  200.,  500., 1000., 2000., 5000.]
            unit: "keV"
        scheme: "ring"
        coordinate: "galactic"
        unit: "cm-2 s-1 sr-1"

        """

        new = cls(nside = parameter['nside'],
                  energy_edges = parameter['energy_edges']['value'] * u.Unit(parameter['energy_edges']['unit']), 
                  scheme = parameter['scheme'], 
                  coordsys = parameter['coordinate'],
                  unit = u.Unit(parameter['unit']))

        return new

    def set_values_from_parameters(self, parameter):
        """
        Set the values accordinng to the specified algorithm. 

        Parameters
        ----------
        parameter : py:class:`yayc.Configurator`
            Parameters for the specified algorithm.

        Notes
        -----
        Currently algorithm_name can be only 'flat'. All of the pixel values in each energy bins will set to the given value.
        parameter should be {'values': [ flux value at 1st energy bin (without unit), flux value at 2nd energy bin, ...]}.

        An example of contents in parameter is like this:

        algorithm: "flat"
        parameter:
            value: [1.0e-2, 1.0e-2, 1.0e-2, 1.0e-2, 1.0e-2]
            unit: "cm-2 s-1 sr-1"

        """

        algorithm_name = parameter['algorithm']
        algorithm_parameter = parameter['parameter']

        if algorithm_name == "flat":
            unit = u.Unit(algorithm_parameter['unit'])
            for idx, value in enumerate(algorithm_parameter['value']):
                self[:,idx] = value * unit
    #    elif algorithm_name == ... 
    #       ...

    def set_values_from_extendedmodel(self, extendedmodel):
        """
        Set the values accordinng to the given astromodels.ExtendedSource.

        Parameters
        ----------
        extendedmodel : :py:class:`astromodels.ExtendedSource`
            Extended source model
        """

        integrated_flux = get_integrated_spectral_model(spectrum = extendedmodel.spectrum.main.shape,
                                                        energy_axis = self.axes[1])
        
        npix = self.axes[0].npix
        coords = self.axes[0].pix2skycoord(np.arange(npix))

        normalized_map = extendedmodel.spatial_shape(coords.l.deg, coords.b.deg) / u.sr

        self[:] = np.tensordot(normalized_map, integrated_flux.contents, axes = 0) 

    def smoothing(self, fwhm = None, sigma = None):
        """
        Smooth a map with a Gaussian filter

        Parameters
        ----------
        fwhm: :py:class:`astropy.units.quantity.Quantity`
            The FWHM of the Gaussian (with a unit of deg or rad). Default: 0 deg
        sigma: :py:class:`astropy.units.quantity.Quantity`
            The sigma of the Gaussian (with a unit of deg or rad). Override fwhm.
        """

        if fwhm is None:
            fwhm = 0.0 * u.deg
        
        if sigma is not None:
            fwhm = 2.354820 * sigma

        allskyimage_new = copy.deepcopy(self)
        
        for i in range(self.axes['Ei'].nbins):
            allskyimage_new[:,i] = hp.smoothing(self[:,i].value, fwhm = fwhm.to('rad').value) * self.unit

        return allskyimage_new



from histpy import Histogram, Axes, Axis

import numpy as np
import astropy.units as u
from astropy.units import Quantity
from scipy import integrate

from threeML import DiracDelta, Constant, Line, Quadratic, Cubic, Quartic, StepFunction, StepFunctionUpper, Cosine_Prior, Uniform_prior, PhAbs, Gaussian

from .functions import get_integrated_spectral_model

class PointSourceResponse(Histogram):
    """
    Handles the multi-dimensional matrix that describes the expected
    response of the instrument for a particular point in the sky.

    Parameters
    ----------
    axes : :py:class:`histpy.Axes`
        Binning information for each variable. The following labels are expected:\n
        - ``Ei``: Real energy
        - ``Em``: Measured energy. Optional
        - ``Phi``: Compton angle. Optional.
        - ``PsiChi``:  Location in the Compton Data Space (HEALPix pixel). Optional.
        - ``SigmaTau``: Electron recoil angle (HEALPix pixel). Optional.
        - ``Dist``: Distance from first interaction. Optional.
    contents : array, :py:class:`astropy.units.Quantity` or :py:class:`sparse.SparseArray`
        Array containing the differential effective area convolved with wht source exposure.
    unit : :py:class:`astropy.units.Unit`, optional
        Physical units, if not specified as part of ``contents``. Units of ``area*time``
        are expected.
    """
    
    @property
    def photon_energy_axis(self):
        """
        Real energy bins (``Ei``).

        Returns
        -------
        :py:class:`histpy.Axes`
        """
        
        return self.axes['Ei']
       
    def get_expectation(self, spectrum):
        """
        Convolve the response with a spectral hypothesis to obtain the expected
        excess counts from the source.

        Parameters
        ----------
        spectrum : :py:class:`threeML.Model`
            Spectral hypothesis.

        Returns
        -------
        :py:class:`histpy.Histogram`
             Histogram with the expected counts on each analysis bin
        """
        
        energy_axis = self.photon_energy_axis

        flux = get_integrated_spectral_model(spectrum, energy_axis)
        
        expectation = np.tensordot(self.contents, flux.contents, axes = ([0], [0]))
        
        # Note: np.tensordot loses unit if we use a sparse matrix as it input.
        if self.is_sparse:
            expectation *= self.unit * flux.unit

        hist = Histogram(self.axes[1:], contents = expectation)
        
        return hist



from abc import ABC, abstractmethod
import astropy.units as u
import numpy as np
import copy
from histpy import Histogram

class ModelBase(Histogram, ABC):
    """
    A base class of the model, i.e., a gamma-ray flux sky, and a gamma-ray distributions in a 3D space.

    Subclasses must override these methods. The `ImageDeconvolution` class will use them in the initialization process.

    Methods:
    - instantiate_from_parameters(cls, parameter)
    - set_values_from_parameters(self, parameter)
    """

    def __init__(self, edges, contents = None, sumw2 = None,
                 labels=None, axis_scale = None, sparse = None, unit = None):

        super().__init__(edges, contents = contents, sumw2 = sumw2, 
                         labels = labels, axis_scale = axis_scale, sparse = sparse, unit = unit)

    @classmethod
    @abstractmethod
    def instantiate_from_parameters(cls, parameter):
        """
        Return an instantiate of the class using given parameters.

        Parameters
        ----------
        parameter : py:class:`yayc.Configurator`
            Parameters for the specified algorithm.

        Returns
        -------
        py:class:`ModelBase`
        """

        raise NotImplementedError

    @abstractmethod
    def set_values_from_parameters(self, parameter):
        """
        Set values accordinng to the give parameters. 

        Parameters
        ----------
        parameter : py:class:`yayc.Configurator`
            Parameters for the specified algorithm.
        """

        raise NotImplementedError

    def mask_pixels(self, mask, fill_value = 0):
        """
        Mask pixels

        Parameters
        ----------
        mask: :py:class:`histpy.histogram.Histogram`
        fill_value: float or :py:class:`astropy.units.quantity.Quantity`
        """

        if not isinstance(fill_value, u.quantity.Quantity) and self.unit is not None:
            fill_value *= self.contents.unit

        model_new = copy.deepcopy(self)
        model_new[:] = np.where(mask.contents, model_new.contents, fill_value)

        return model_new



import logging
logger = logging.getLogger(__name__)

import pandas as pd
from tqdm.autonotebook import tqdm
import numpy as np
import healpy as hp
from astropy.io import fits
import astropy.units as u

from cosipy.spacecraftfile import SpacecraftAttitudeMap

class SpacecraftAttitudeExposureTable(pd.DataFrame):
    """
    A class to analyze exposure time per each spacecraft attitude
    
    Table columns are:
    - scatt_binning_index: int 
    - healpix_index: list of tuple. Each tuple is (healpix_index_zpointing, healpix_index_xpointing).
    - zpointing: np.array of [l, b] in degrees. Array of z-pointings assigned to each scatt bin.
    - xpointing: np.array of [l, b] in degrees. Array of x-pointings assigned to each scatt bin.
    - zpointing_averaged: [l, b] in degrees. Averaged z-pointing in each scatt bin.
    - xpointing_averaged: [l, b] in degrees. Averaged x-pointing in each scatt bin.
    - delta_time: np.array of float in second. Exposure times for pointings assigned to each scatt bin.
    - exposure: float in second. total exposure for each scatt bin.
    - num_pointings: number of pointings assigned to each scatt bin.
    - bkg_group: index of the backgroud group. will be used in the data analysis.

    Attributes
    ----------
    df : :py:class:`pd.DataFrame`
        pandas dataframe with the above columns
    nside : int
        Healpix NSIDE parameter.
    scheme : str, default 'ring'
        Healpix scheme. Either 'ring', 'nested'. 
    """

    def __init__(self, df, nside, scheme = 'ring'):

        super().__init__(pd.DataFrame(df))

        self.nside = nside

        if scheme == 'ring' or scheme == 'nested':
            self.scheme = scheme
        else:
            logger.warning('The scheme should be "ring" or "nested" in SpacecraftAttitudeExposureTable. It will be set to "ring".')
            self.scheme = 'ring'

    def __eq__(self, other):
        for name in ['scatt_binning_index', 'healpix_index', 'exposure', 'num_pointings', 'bkg_group']:
            if not np.all(self[name] == other[name]):
                return False
        
        for name in ['delta_time', 'zpointing', 'xpointing', 'zpointing_averaged', 'xpointing_averaged']:
            for self_, other_ in zip(self[name], other[name]):
                if not np.all(self_ == other_):
                    return False

        return (self.nside == other.nside) and (self.scheme == other.scheme)

    @classmethod
    def from_orientation(cls, orientation, nside, scheme = 'ring', start = None, stop = None, min_exposure = None, min_num_pointings = None):
        """
        Produce exposure table from orientation.

        Parameters
        ----------
        orientation : :py:class:`cosipy.spacecraftfile.SpacecraftFile` 
            Orientation
        nside : int
            Healpix NSIDE parameter.
        scheme : str, default 'ring'
            Healpix scheme. Either 'ring', 'nested'.
        start : :py:class:`astropy.time.Time` or None, default None
            Start time to analyze the orientation
        stop : :py:class:`astropy.time.Time` or None, default None
            Stop time to analyze the orientation
        min_exposure : float or None, default None
            Minimum exposure time required for each scatt bin
        min_num_pointings : int or None, default None
            Minimum number of pointings required for each scatt bin

        Returns
        -------
        :py:class:`cosipy.spacecraftfile.SpacecraftAttitudeExposureTable`
        """
        
        df = cls.analyze_orientation(orientation, nside, scheme, start, stop, min_exposure, min_num_pointings)

        new = cls(df, nside, scheme)

        return new

    # GTI should be a mandary parameter
    @classmethod
    def analyze_orientation(cls, orientation, nside, scheme = 'ring', start = None, stop = None, min_exposure = None, min_num_pointings = None):
        """
        Produce pd.DataFrame from orientation.

        Parameters
        ----------
        orientation : :py:class:`cosipy.spacecraftfile.SpacecraftFile` 
            Orientation
        nside : int
            Healpix NSIDE parameter.
        scheme : str, default 'ring'
            Healpix scheme. Either 'ring', 'nested'.
        start : :py:class:`astropy.time.Time` or None, default None
            Start time to analyze the orientation
        stop : :py:class:`astropy.time.Time` or None, default None
            Stop time to analyze the orientation
        min_exposure : float or None, default None
            Minimum exposure time required for each scatt bin
        min_num_pointings : int or None, default None
            Minimum number of pointings required for each scatt bin

        Returns
        -------
        :py:class:`pd.DataFrame`
        """

        logger.info("angular resolution: ", hp.nside2resol(nside) * 180 / np.pi, "deg.")    
    
        indices_healpix = [] # (idx_z, idx_x)
        delta_times = []
        xpointings = [] # [l_x, b_x]
        zpointings = [] # [l_z, b_z]
                
        if start is not None and stop is not None:
            orientation = orientation.source_interval(start, stop)
        elif start is not None:
            logger.error("please specify the stop time")
            return
        elif stop is not None:
            logger.error("please specify the start time")
            return
        
        ori_time = orientation.get_time()
            
        logger.info("duration: ", (ori_time[-1] - ori_time[0]).to("day"))
        
        attitude = orientation.get_attitude()
        
        pointing_list = attitude.transform_to("galactic").as_axes()

        n_pointing = len(pointing_list[0])
        
        x_1, x_2 = pointing_list[0][:-1], pointing_list[0][1:]
        l_x, b_x =  0.5 * (x_1.l.degree + x_2.l.degree), 0.5 * (x_1.b.degree + x_2.b.degree)

        z_1, z_2 = pointing_list[2][:-1], pointing_list[2][1:]
        l_z, b_z =  0.5 * (z_1.l.degree + z_2.l.degree), 0.5 * (z_1.b.degree + z_2.b.degree)

        if scheme == 'ring':
            nest = False
        elif scheme == 'nested':
            nest = True
        else:
            logger.warning('Warning: the scheme should be "ring" or "nested". It was set to "ring".')
            nest = False

        idx_x = hp.ang2pix(nside, l_x, b_x, nest=nest, lonlat=True)
        idx_z = hp.ang2pix(nside, l_z, b_z, nest=nest, lonlat=True)
        
        delta_time = (ori_time[1:] - ori_time[:-1]).to('s').value
        
        for i in tqdm(range(n_pointing - 1)):
            
            if (idx_z[i], idx_x[i]) in indices_healpix:
                idx = indices_healpix.index((idx_z[i], idx_x[i]))
                delta_times[idx].append(delta_time[i])
                xpointings[idx].append([l_x[i], b_x[i]])
                zpointings[idx].append([l_z[i], b_z[i]])            
                
            else:
                indices_healpix.append((idx_z[i], idx_x[i]))
                delta_times.append([delta_time[i]])
                xpointings.append([[l_x[i], b_x[i]]])
                zpointings.append([[l_z[i], b_z[i]]])
        
        indices_scatt_binning = [i for i in range(len(indices_healpix))] 
        
        # to numpy
        zpointings = [ np.array(_) for _ in zpointings]
        xpointings = [ np.array(_) for _ in xpointings]

        zpointings_averaged = [ cls._get_averaged_pointing(z, dt) for (z, dt) in zip(zpointings, delta_times) ]
        xpointings_averaged = [ cls._get_averaged_pointing(x, dt) for (x, dt) in zip(xpointings, delta_times) ]

        exposures = [ np.sum(np.array(_)) for _ in delta_times]
        num_pointings = [ len(_) for _ in delta_times]
        bkg_groups = [ 0 for i in delta_times]
        
        df = pd.DataFrame(data = {'scatt_binning_index': indices_scatt_binning, 'healpix_index': indices_healpix, 
                                  'zpointing': zpointings,
                                  'xpointing': xpointings,
                                  'zpointing_averaged': zpointings_averaged,
                                  'xpointing_averaged': xpointings_averaged,
                                  'delta_time': delta_times, 
                                  'exposure': exposures,
                                  'num_pointings': num_pointings,
                                  'bkg_group': bkg_groups})
        
        if min_exposure is not None:
            df = df[df['exposure'] >= min_exposure]

        if min_num_pointings is not None:
            df = df[df['num_pointings'] >= min_num_pointings]
        
        if min_exposure is not None or min_num_pointings is not None:
            df['scatt_binning_index'] = [i for i in range(len(df))] 
        
        return df 

    @classmethod
    def from_fits(cls, filename):
        """
        Read exposure table from a fits file.

        Parameters
        ----------
        filename : str
            Path to file
        
        Returns
        -------
        :py:class:`cosipy.image_deconvolution.SpacecraftAttitudeExposureTable`
        """

        infile = fits.open(filename)
        hdu = infile[1]
    
        if hdu.name != "EXPOSURETABLE":
            logger.error("cannot find EXPOSURETABLE")
            return
    
        indices_scatt_binning = hdu.data['scatt_binning_index']
        indices_healpix = [ (z, x) for (z, x) in zip(hdu.data['healpix_index_z_pointing'], hdu.data['healpix_index_x_pointing']) ]

        zpointings = [ [ [l, b] for (l, b) in zip(z_l, z_b) ] for (z_l, z_b) in zip(hdu.data['zpointing_l'], hdu.data['zpointing_b']) ]
        zpointings = [ np.array(_) for _ in zpointings]

        xpointings = [ [ [l, b] for (l, b) in zip(x_l, x_b) ] for (x_l, x_b) in zip(hdu.data['xpointing_l'], hdu.data['xpointing_b']) ]
        xpointings = [ np.array(_) for _ in xpointings]

        zpointings_averaged = [ np.array([z_ave_l, z_ave_b]) for (z_ave_l, z_ave_b)  in zip(hdu.data['zpointing_averaged_l'], hdu.data['zpointing_averaged_b']) ]

        xpointings_averaged = [ np.array([x_ave_l, x_ave_b]) for (x_ave_l, x_ave_b)  in zip(hdu.data['xpointing_averaged_l'], hdu.data['xpointing_averaged_b']) ]

        delta_times = np.array(hdu.data['delta_time'])

        exposures = hdu.data['exposure']

        num_pointings = hdu.data['num_pointings']

        bkg_groups = hdu.data['bkg_group']
        
        df = pd.DataFrame(data = {'scatt_binning_index': indices_scatt_binning, 'healpix_index': indices_healpix, 
                                  'zpointing': zpointings,
                                  'xpointing': xpointings,
                                  'zpointing_averaged': zpointings_averaged,
                                  'xpointing_averaged': xpointings_averaged,
                                  'delta_time': delta_times, 
                                  'exposure': exposures,
                                  'num_pointings': num_pointings,
                                  'bkg_group': bkg_groups})

        nside = hdu.header['NSIDE']
        scheme = hdu.header['SCHEME']
        
        new = cls(df, nside, scheme)

        return new

    def save_as_fits(self, filename, overwrite = False):
        """
        Save exposure table as a fits file.

        Parameters
        ----------
        filename : str
            Path to file
        overwrite : bool, default False
        """

        # primary HDU
        primary_hdu = fits.PrimaryHDU()
    
        #exposure table
        names = ['scatt_binning_index', 'exposure', 'num_pointings', 'bkg_group']
        formats = ['K', 'D', 'K', 'K']
        units = ['', 's', '', '']
        
        columns = [ fits.Column(name=names[i], array=self[names[i]].to_numpy(), format = formats[i], unit = units[i]) 
                     for i in range(len(names))]
        
        column_healpix_index_z_pointing = fits.Column(name='healpix_index_z_pointing', 
                                                      array=np.array([idx[0] for idx in self['healpix_index']]), format = 'K')
        column_healpix_index_x_pointing = fits.Column(name='healpix_index_x_pointing', 
                                                      array=np.array([idx[1] for idx in self['healpix_index']]), format = 'K')
        
        columns.append(column_healpix_index_z_pointing)
        columns.append(column_healpix_index_x_pointing)
    
        column_delta_time = fits.Column(name='delta_time', format='PD()', unit = 's',
                                        array=np.array(self['delta_time'].array, dtype=np.object_))
        columns.append(column_delta_time)    
        
        column_zpointing_l = fits.Column(name='zpointing_l', format='PD()', unit = 'degree',
                                        array=np.array([[pointing[0] for pointing in pointings] for pointings in self['zpointing']], dtype=np.object_))
        columns.append(column_zpointing_l)    
    
        column_zpointing_b = fits.Column(name='zpointing_b', format='PD()', unit = 'degree',
                                        array=np.array([[pointing[1] for pointing in pointings] for pointings in self['zpointing']], dtype=np.object_))
        columns.append(column_zpointing_b)   
    
        column_xpointing_l = fits.Column(name='xpointing_l', format='PD()', unit = 'degree',
                                        array=np.array([[pointing[0] for pointing in pointings] for pointings in self['xpointing']], dtype=np.object_))
        columns.append(column_xpointing_l)    
    
        column_xpointing_b = fits.Column(name='xpointing_b', format='PD()', unit = 'degree',
                                        array=np.array([[pointing[1] for pointing in pointings] for pointings in self['xpointing']], dtype=np.object_))
        columns.append(column_xpointing_b)  

        column_zpointing_averaged_l = fits.Column(name='zpointing_averaged_l', format='D', unit = 'degree',
                                                  array=np.array([_[0] for _ in self['zpointing_averaged']]))
        columns.append(column_zpointing_averaged_l)    

        column_zpointing_averaged_b = fits.Column(name='zpointing_averaged_b', format='D', unit = 'degree',
                                                  array=np.array([_[1] for _ in self['zpointing_averaged']]))
        columns.append(column_zpointing_averaged_b)    

        column_xpointing_averaged_l = fits.Column(name='xpointing_averaged_l', format='D', unit = 'degree',
                                                  array=np.array([_[0] for _ in self['xpointing_averaged']]))
        columns.append(column_xpointing_averaged_l)    

        column_xpointing_averaged_b = fits.Column(name='xpointing_averaged_b', format='D', unit = 'degree',
                                                  array=np.array([_[1] for _ in self['xpointing_averaged']]))
        columns.append(column_xpointing_averaged_b)    
        
        table_hdu = fits.BinTableHDU.from_columns(columns) 
        table_hdu.name = 'exposuretable'

        table_hdu.header['nside'] = self.nside
        table_hdu.header['scheme'] = self.scheme
        
        #save file    
        hdul = fits.HDUList([primary_hdu, table_hdu])    
        hdul.writeto(filename, overwrite = overwrite)

    @classmethod
    def _get_averaged_pointing(cls, pointing, delta_time):
        """
        Calculate an averaged pointing from given lists of pointings and exposure time on each pointing

        Parameters
        ----------
        pointing : list of np.array
            List of pointings in degrees, e.g., [ np.array([l, b]), np.array([l, b]), ...]
        delta_time : list of float
            List of exposure time in seconds for each pointing, e.g, [ 1.0, 1.0, ...] 

        Returns
        -------
        :py:class:`np.array`
            Averaged pointing in degrees, as np.array([l, b])
        """

        averaged_vector = np.sum(hp.ang2vec(pointing.T[0], pointing.T[1], lonlat = True).T * delta_time, axis = (1))
        averaged_vector /= np.linalg.norm(averaged_vector)

        averaged_l = hp.vec2ang(averaged_vector, lonlat = True)[0][0]
        averaged_b = hp.vec2ang(averaged_vector, lonlat = True)[1][0]
    
        averaged_pointing = np.array([averaged_l, averaged_b])

        return averaged_pointing 

    def calc_pointing_trajectory_map(self):
        """
        Calculate a 2-dimensional map showing exposure time for each spacecraft attitude.

        Returns
        -------
        :py:class:`cosipy.spacecraft.SpacecraftAttitudeMap`

        Notes
        -----
        The default axes in SpacecraftAttitudeMap is x- and y-pointings,
        but here the spacecraft attitude is described with z- and x-pointings. 
        """
    
        map_pointing_zx = SpacecraftAttitudeMap(nside = self.nside, scheme = self.scheme, coordsys = 'galactic', labels = ['z', 'x'])
    
        for hp_index, exposure in zip(self['healpix_index'], self['exposure']):
            map_pointing_zx[hp_index[0], hp_index[1]] = exposure * u.s
        
        return map_pointing_zx



# Import
from cosipy.data_io import UnBinnedData 
import matplotlib.pyplot as plt
import numpy as np
import sys
import pandas as pd
import logging
logger = logging.getLogger(__name__)

try:
    # Load MEGAlib into ROOT
    import ROOT as M
    M.gSystem.Load("$(MEGAlib)/lib/libMEGAlib.so")

    # Initialize MEGAlib
    G = M.MGlobal()
    G.Initialize()
    
except:
    pass

class ReadTraTest(UnBinnedData):

    """Old method for reading tra file, used for unit testing."""
    
    def read_tra_old(self, output_name, make_plots=True):
        
        """Reads in MEGAlib .tra (or .tra.gz) file.
       
        This method uses MEGAlib to read events from the tra file. 
        This is used to compare to the new event reader, which is 
        independent of MEGAlib. 
        

        Parameters
        ----------
        output_name : str
            Prefix of output file.
        make_plots : bool, optional
            Option to make binning plot.

        Returns
        -------
        cosi_dataset : dict
            Returns COSI dataset as a dictionary of the form:
            cosi_dataset = {'Full filename':self.data_file,\
                        'Energies':erg,\
                        'TimeTags':tt,\
                        'Xpointings':np.array([lonX,latX]).T,\
                        'Ypointings':np.array([lonY,latY]).T,\
                        'Zpointings':np.array([lonZ,latZ]).T,\
                        'Phi':phi,\
                        'Chi local':chi_loc,\
                        'Psi local':psi_loc,\
                        'Distance':dist,\
                        'Chi galactic':chi_gal,\
                        'Psi galactic':psi_gal}
        
        Note
        ----
        This method sets the instance attribute, cosi_dataset, 
        but it does not explicitly return this.  
        """

        # tra file to use:
        tra_file = self.data_file

        # Log message:
        logger.info("Read tra test...")
        
         
        # Check if file exists:
        Reader = M.MFileEventsTra()
        if Reader.Open(M.MString(tra_file)) == False:
            logger.error("Unable to open file %s. Aborting!" %self.data_file)
            sys.exit()

        # Initialise empty lists:
            
        # Total photon energy
        erg = []
        # Time tag in UNIX time
        tt = []
        # Event Type (0: CE; 4:PE; ...)
        et = []
        # Latitude of X direction of spacecraft
        latX = []
        # Lontitude of X direction of spacecraft
        lonX = []
        # Latitude of Z direction of spacecraft
        latZ = []
        # Longitude of Z direction of spacecraft
        lonZ = []
        # Compton scattering angle
        phi = []
        # Measured data space angle chi (azimuth direction; 0..360 deg)
        chi_loc = []
        # Measured data space angle psi (polar direction; 0..180 deg)
        psi_loc = []
        # First lever arm distance in cm
        dist = []
        # Measured gal angle chi (lon direction)
        chi_gal = []
        # Measured gal angle psi (lat direction)
        psi_gal = [] 

        # Browse through tra file, select events, and sort into corresponding list:
        # Note: The Reader class from MEGAlib knows where an event starts and ends and
        # returns the Event object which includes all information of an event.
        # Note: Here only select Compton events (will add Photo events later as optional).
        # Note: All calculations and definitions taken from:
        # /MEGAlib/src/response/src/MResponseImagingBinnedMode.cxx.

        while True:
            
            Event = Reader.GetNextEvent()
            if not Event:
                break
                
            # Total Energy:
            erg.append(Event.Ei())
            # Time tag in UNIX seconds:
            tt.append(Event.GetTime().GetAsSeconds())
            # Event type (0 = Compton, 4 = Photo):
            et.append(Event.GetEventType())
            # x axis of space craft pointing at GAL latitude:
            latX.append(Event.GetGalacticPointingXAxisLatitude())
            # x axis of space craft pointing at GAL longitude:
            lonX.append(Event.GetGalacticPointingXAxisLongitude())
            # z axis of space craft pointing at GAL latitude:
            latZ.append(Event.GetGalacticPointingZAxisLatitude())
            # z axis of space craft pointing at GAL longitude:
            lonZ.append(Event.GetGalacticPointingZAxisLongitude())    
            # Compton scattering angle:
            phi.append(Event.Phi()) 
            # Data space angle chi (azimuth):
            chi_loc.append((-Event.Dg()).Phi())
            # Data space angle psi (polar):
            psi_loc.append((-Event.Dg()).Theta())
            # Interaction length between first and second scatter in cm:
            dist.append(Event.FirstLeverArm())
            # Gal longitude angle corresponding to chi:
            chi_gal.append((Event.GetGalacticPointingRotationMatrix()*Event.Dg()).Phi())
            # Gal longitude angle corresponding to chi:
            psi_gal.append((Event.GetGalacticPointingRotationMatrix()*Event.Dg()).Theta())
                
        # Initialize arrays:
        erg = np.array(erg)
        tt = np.array(tt)
        et = np.array(et)

        latX = np.array(latX)
        lonX = np.array(lonX)
        # Change longitudes to from 0..360 deg to -180..180 deg
        lonX[lonX > np.pi] -= 2*np.pi

        latZ = np.array(latZ)
        lonZ = np.array(lonZ)
        # Change longitudes to from 0..360 deg to -180..180 deg
        lonZ[lonZ > np.pi] -= 2*np.pi

        phi = np.array(phi)

        chi_loc = np.array(chi_loc)
        self.chi_loc_old = chi_loc

        # Change azimuth angle to 0..360 deg
        chi_loc[chi_loc < 0] += 2*np.pi

        psi_loc = np.array(psi_loc)
        self.psi_loc_old = psi_loc
        
        # For comparing chi_loc, psi_loc=0 values are arbitrary,
        # so we exclude them from the comparison. 
        psi_zero_index = psi_loc == 0

        dist = np.array(dist)

        chi_gal = np.array(chi_gal)
        psi_gal = np.array(psi_gal)
        self.chi_gal_old = chi_gal
        self.psi_gal_old = psi_gal
         
        # Construct Y direction from X and Z direction
        lonlatY = self.construct_scy(np.rad2deg(lonX),np.rad2deg(latX),
                                np.rad2deg(lonZ),np.rad2deg(latZ))
        lonY = np.deg2rad(lonlatY[0])
        latY = np.deg2rad(lonlatY[1])

        # Avoid negative zeros
        chi_loc[np.where(chi_loc == 0.0)] = np.abs(chi_loc[np.where(chi_loc == 0.0)])
       
        # Make observation dictionary
        cosi_dataset = {'Energies':erg,
                        'TimeTags':tt,
                        'Xpointings':np.array([lonX,latX]).T,
                        'Ypointings':np.array([lonY,latY]).T,
                        'Zpointings':np.array([lonZ,latZ]).T,
                        'Phi':phi,
                        'Chi local':self.chi_loc_old,
                        'Psi local':self.psi_loc_old,
                        'Distance':dist,
                        'Chi galactic':self.chi_gal_old,
                        'Psi galactic':self.psi_gal_old} 
        self.cosi_dataset = cosi_dataset

        # Write unbinned data to file (either fits or hdf5):
        self.write_unbinned_output(output_name) 
        
        return 




from histpy import Histogram, HealpixAxis

import astropy.units as u

class SpacecraftAttitudeMap(Histogram):

    def __init__(self,
                 nside,
                 scheme = 'ring',
                 coordsys = 'galactic',
                 labels = ['x', 'y']
                 ):
        """
        Bin the spacecraft attitude history into a 4D histogram that contains 
        the accumulated time the axes of the spacecraft where looking at a 
        given direction. 

        Same arguments as an HealpixAxis.

        Parameters
        ----------
        nside : int
            The nside of the spacecraft attitude map.
        scheme : str, optional
            The scheme of the spacecraft attitude map (the default is "ring").
        coordsys : str, optional
            The coordinate system of the spacecraft attitude map (the default is "galactic").
        labels : list, optional
            The labels of the two axes of the spacecraft attitude map (the default is `["x", "y"]`.

            
        """

        super().__init__([HealpixAxis(nside = nside,
                                      scheme = scheme,
                                      coordsys = coordsys,
                                      label = labels[0]), 
                          HealpixAxis(nside = nside,
                                      scheme = scheme,
                                      coordsys = coordsys,
                                      label = labels[1])],
                         sparse = True,
                         unit = u.s)    

        



import logging
logger = logging.getLogger(__name__)

from histpy import Histogram, Axis, Axes
import astropy.units as u
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy import integrate
from iminuit import Minuit

class LineBackgroundEstimation:
    """
    A class for estimating and modeling background in line spectra.

    This class provides methods for setting up a background model,
    fitting it to data, and generating background model histograms.

    Attributes
    ----------
    event_histogram : Histogram
        The input event histogram.
    energy_axis : Axis
        The energy axis of the event histogram.
    energy_spectrum : Histogram
        The projected energy spectrum.
    bkg_spectrum_model : callable
        The background spectrum model function.
    bkg_spectrum_model_parameter : list
        The parameters of the background spectrum model.
    mask : ndarray
        Boolean mask for excluding regions from the fit.
    """

    def __init__(self, event_histogram):
        """
        Initialize the LineBackgroundEstimation object.

        Parameters
        ----------
        event_histogram : Histogram
            The input event histogram.
        """
        # event histogram
        self.event_histogram = event_histogram

        # projected histogram onto the energy axis
        self.energy_axis = self.event_histogram.axes['Em']
        self.energy_spectrum = self.event_histogram.project('Em')
        if self.energy_spectrum.is_sparse:
            self.energy_spectrum = self.energy_spectrum.to_dense()
        
        self.energy_spectrum.clear_underflow_and_overflow()

        # background fitting model
        self.bkg_spectrum_model = None
        self.bkg_spectrum_model_parameter = None

        # bins to be masked
        self.mask = np.zeros(self.energy_axis.nbins, dtype=bool)
        
    def set_bkg_energy_spectrum_model(self, bkg_spectrum_model, bkg_spectrum_model_parameter):
        """
        Set the background energy spectrum model and its initial parameters.

        Parameters
        ----------
        bkg_spectrum_model : callable
            The background spectrum model function.
        bkg_spectrum_model_parameter : list
            Initial parameters for the background spectrum model.
        """
        self.bkg_spectrum_model = bkg_spectrum_model
        self.bkg_spectrum_model_parameter = bkg_spectrum_model_parameter

    def set_mask(self, *mask_energy_ranges):
        """
        Set mask for excluding energy ranges from the fit.

        Parameters
        ----------
        *mask_energy_ranges : tuple
            Variable number of energy range tuples to be masked.
        """
        self.mask = np.zeros(self.energy_axis.nbins, dtype=bool)
        for mask_energy_range in mask_energy_ranges:
            this_mask = (mask_energy_range[0] <= self.energy_axis.bounds[:, 1]) & (self.energy_axis.bounds[:, 0] <= mask_energy_range[1])
            self.mask = self.mask | this_mask
    
    def _calc_expected_spectrum(self, *args):
        """
        Calculate the expected spectrum based on the current model and parameters.

        Parameters
        ----------
        *args : float
            Model parameters.

        Returns
        -------
        ndarray
            The calculated expected spectrum.
        """
        return np.array([integrate.quad(lambda x: self.bkg_spectrum_model(x, *args), *energy_range)[0] for energy_range in self.energy_axis.bounds.value])

    def _negative_log_likelihood(self, *args):
        """
        Calculate the negative log-likelihood for the current model and parameters.

        Parameters
        ----------
        *args : float
            Model parameters.

        Returns
        -------
        float
            The calculated negative log-likelihood.
        """
        expected_spectrum = self._calc_expected_spectrum(*args)
        return -np.sum(self.energy_spectrum.contents[~self.mask] * np.log(expected_spectrum)[~self.mask]) + np.sum(expected_spectrum[~self.mask])
    
    def plot_energy_spectrum(self):
        """
        Plot the energy spectrum and the fitted model if available.

        Returns
        -------
        tuple
            A tuple containing the matplotlib axis object and any additional objects returned by the plotting function.
        """
        ax, _ = self.energy_spectrum.draw(label='input data')

        # plot background model
        if self.bkg_spectrum_model is not None:
            expected_spectrum = self._calc_expected_spectrum(*self.bkg_spectrum_model_parameter)
            ax.plot(self.energy_axis.centers, expected_spectrum, label='model')

        # shade mask regions
        start, end = None, None
        for i, this_mask in enumerate(self.mask):
            if this_mask:
                if start is None:
                    start, end = self.energy_axis.bounds[i]
                else:
                    _, end = self.energy_axis.bounds[i]
            else:
                if start is not None:
                    ax.axvspan(start.value, end.value, color='lightgrey', alpha=0.5)
                    start, end = None, None
        
        if start is not None:
            ax.axvspan(start.value, end.value, color='lightgrey', alpha=0.5)

        # legend and grid
        ax.legend()
        ax.grid()
        
        return ax, _
        
    def fit_energy_spectrum(self):
        """
        Fit the background energy spectrum model to the data.

        Returns
        -------
        Minuit
            The Minuit object containing the fit results.
        """
        m = Minuit(self._negative_log_likelihood, *self.bkg_spectrum_model_parameter)
        m.errordef = Minuit.LIKELIHOOD
        
        m.migrad()
        m.hesse()

        # update the background model parameters
        self.bkg_spectrum_model_parameter = list(m.values)
        self.bkg_spectrum_model_parameter_errors = list(m.errors)
        
        return m

    def _get_weight_indices(self, energy_range):
        """
        Get the weight and indices for a given energy range.

        Parameters
        ----------
        energy_range : tuple
            The energy range to calculate the weight for.

        Returns
        -------
        tuple
            A tuple containing the calculated weight and the corresponding energy indices.
        """
        energy_indices = np.where((energy_range[0] <= self.energy_axis.lower_bounds) & (self.energy_axis.upper_bounds <= energy_range[1]))[0]

        if len(energy_indices) == 0:
            raise ValueError("The input energy range is too narrow to find a corresponding energy bin.")

        integrate_energy_range = [self.energy_axis.lower_bounds[energy_indices[0]].value, self.energy_axis.upper_bounds[energy_indices[-1]].value]

        if integrate_energy_range[0] != energy_range[0].value or integrate_energy_range[1] != energy_range[1].value:
            logger.info(f"The energy range {energy_range.value} is modified to {integrate_energy_range}")
        weight = integrate.quad(lambda x: self.bkg_spectrum_model(x, *self.bkg_spectrum_model_parameter), *integrate_energy_range)[0]
        return weight, energy_indices

    def generate_bkg_model_histogram(self, source_energy_range, bkg_estimation_energy_ranges):
        """
        Generate a background model histogram based on the fitted model.

        Parameters
        ----------
        bkg_estimation_energy_ranges : list of tuple
            List of energy ranges for background estimation.
        smoothing_fwhm : float, optional
            Full width at half maximum for smoothing, by default None.

        Returns
        -------
        Histogram
            The generated background model histogram.
        """
        # intergrated spectrum in the background estimation energy ranges
        weights = []
        energy_indices_list = []
        for bkg_estimation_energy_range in bkg_estimation_energy_ranges:
            weight, energy_indices = self._get_weight_indices(bkg_estimation_energy_range)
            weights.append(weight)
            energy_indices_list.append(energy_indices)
            
        # intergrated spectrum in the source region
        source_weight = integrate.quad(lambda x: self.bkg_spectrum_model(x, *self.bkg_spectrum_model_parameter), *source_energy_range.value)[0]

        # prepare a new histogram
        new_axes = []
        for axis in self.event_histogram.axes:
            if axis.label != "Em":
                new_axes.append(axis)
            else:
                new_axes.append(Axis(source_energy_range, label = "Em"))
        
        bkg_model_histogram = Histogram(new_axes)

        # fill contents
        for energy_indices in energy_indices_list:
            for energy_index in energy_indices:
                if new_axes[0].label != "Em":
                    bkg_model_histogram[:] += self.event_histogram[:,energy_index].todense()
                else:
                    bkg_model_histogram[:] += self.event_histogram[energy_index].todense()

        # normalization
        corr_factor = source_weight / np.sum(weights)
        bkg_model_histogram[:] *= corr_factor

        return bkg_model_histogram



# Imports:
import numpy as np
from astropy.table import Table
from astropy.io import fits
from scipy import interpolate
import h5py
import time
import cosipy
from cosipy.data_io import DataIO
from cosipy.spacecraftfile import SpacecraftFile
import gzip
import astropy.coordinates as astro_co
import astropy.units as u
from astropy.coordinates import SkyCoord
from scoords import Attitude
from scoords import SpacecraftFrame
import logging
import sys
import math
from tqdm import tqdm
import subprocess
import gc
import os
import time
logger = logging.getLogger(__name__)


class UnBinnedData(DataIO):
    """Handles unbinned data."""

    def read_tra(self, output_name=None, run_test=False, use_ori=False,
            event_min=None, event_max=None):
        
        """Reads MEGAlib .tra (or .tra.gz) file and creates cosi datset.
        
        Parameters
        ----------
        output_name : str, optional
            Prefix of output file (default is None, in which case no 
            output is written). 
        run_test : bool, optional 
            This is for unit testing only! Keep False unless 
            comparing to MEGAlib calculations. 
        use_ori : bool, optional
            Option to get pointing information from the orientation 
            file, based on event time-stamps (default is False, in 
            which case the pointing information comes from the event 
            file itself). Note: this is an option for now, but will 
            later be the default. 
        event_min : int, optional
            Minimum event number to process (inclusive). All events 
            below this will be skipped.        
        event_max : int, optional
            Maximum event number to process (non-inclusive). All 
            events at and above this will be skipped. 
            
            Note: event_min and event_max correspond to the total 
            number of events in the file, which is not necessarily the 
            same as the event ID number. The purpose of this is to 
            allow the data to be read in chunks, in order to overcome 
            memory limitations, depending on the user's system.

        Returns
        -------
        cosi_dataset, dict
            The returned dictionary contains the COSI dataset, which
            has the form:
            cosi_dataset = {'Energies':erg,\
                        'TimeTags':tt,\
                        'Xpointings':np.array([lonX,latX]).T,\
                        'Ypointings':np.array([lonY,latY]).T,\
                        'Zpointings':np.array([lonZ,latZ]).T,\
                        'Phi':phi,\
                        'Chi local':chi_loc,\
                        'Psi local':psi_loc,\
                        'Distance':dist,\
                        'Chi galactic':chi_gal,\
                        'Psi galactic':psi_gal}\
            Arrays contain unbinned photon data. 
        
        Notes
        -----
        The current code is only able to handle data with Compton 
        events. It will need to be modified to handle single-site 
        and pair events. 

        This method sets the instance attribute, cosi_dataset, 
        but it does not explicitly return this.  
        """
   
        start_time = time.time()

        # Initialise empty lists:
            
        # Total photon energy
        erg = []
        # Time tag in UNIX time
        tt = []
        # Event Type (CE or PE)
        et = []
        # Galactic latitude of X direction of spacecraft
        latX = []
        # Galactic lontitude of X direction of spacecraft
        lonX = []
        # Galactic latitude of Z direction of spacecraft
        latZ = []
        # Galactic longitude of Z direction of spacecraft
        lonZ = [] 
        # Compton scattering angle
        phi = []
        # Measured data space angle chi (azimuth direction; 0..360 deg)
        chi_loc = []
        # Measured data space angle psi (polar direction; 0..180 deg)
        psi_loc = []
        # Measured gal angle chi (lon direction)
        chi_gal = []
        # Measured gal angle psi (lat direction)
        psi_gal = []
        # Components of dg (position vector from 1st interaion to 2nd)
        dg_x = []
        dg_y = []
        dg_z = []

        # Define electron rest energy, which is used in calculation
        # of Compton scattering angle.
        c_E0 = 510.9989500015 # keV

        # This is for unit testing purposes only.
        # Use same value as MEGAlib for direct comparison: 
        if run_test == True:
            c_E0 = 510.999

        logger.info("Preparing to read file...")

        # Open .tra.gz file:
        if self.data_file.endswith(".gz"):
            f = gzip.open(self.data_file,"rt")
            
            # Need to get number of lines for progress bar.
            # First try fast method for unix-based systems:
            try:
                proc=subprocess.Popen('gunzip -c %s | wc -l' %self.data_file, \
                        shell=True, stdout=subprocess.PIPE)
                num_lines = float(proc.communicate()[0])

            # If fast method fails, use long method, which should work in all cases.
            except:
                logger.info("Initial attempt failed.")
                logger.info("Using long method...")
                g = gzip.open(self.data_file,"rt")
                num_lines = sum(1 for line in g)
                g.close()
    
        # Open .tra file:
        elif self.data_file.endswith(".tra"):
            f = open(self.data_file,"r")

            try:
                proc=subprocess.Popen('wc -l < %s' %self.data_file, \
                        shell=True, stdout=subprocess.PIPE)
                num_lines = float(proc.communicate()[0])
                
            except:
                logger.info("Initial attempt failed.")
                logger.info("Using long method...")
                g = open(self.data_file,"rt")
                num_lines = sum(1 for line in g)
                g.close()

        else: 
            logger.error("ERROR: Input data file must have '.tra' or '.gz' extenstion.")
            sys.exit()
        
        
        # Read tra file line by line:
        logger.info("Reading file...")
        N_events = 0 # number of events
        pbar = tqdm(total=num_lines) # start progress bar
        for line in f:
         
            this_line = line.strip().split()
            pbar.update(1) # update progress bar

            # Make sure line isn't empty:
            if len(this_line) == 0:
                continue
                
            # Event type: 
            if this_line[0] == "ET":
                # Check that we are looking at CO events
                if this_line[1] == "CO":
                    et.append(this_line[1])
                else: 
                    raise ValueError(f"Error: Expected CO event, but got '{this_line[1]}'.")
            
            # Count the number of events:
            if this_line[0] == "ID":
                N_events += 1
                
            # Option to only parse a subset of events:
            if event_min != None:
                if N_events < event_min:
                    continue
            if event_max != None:
                if N_events >= event_max:
                    pbar.close()
                    logger.info("Stopping here: only reading a subset of events")
                    break

            # Total photon energy and Compton angle: 
            if this_line[0] == "CE":

                # Compute the total photon energy:
                m_Eg = float(this_line[1]) # Energy of scattered gamma ray in keV
                m_Ee = float(this_line[3]) # Energy of recoil electron in keV
                this_erg = m_Eg + m_Ee
                erg.append(this_erg) 
             
                # Compute the Compton scatter angle due to the standard equation,
                # i.e. neglect the movement of the electron,
                # which would lead to a Doppler-broadening.
                this_value = 1.0 - c_E0 * (1.0/m_Eg - 1.0/(m_Ee + m_Eg))
                this_phi = np.arccos(this_value) # radians
                phi.append(this_phi)
                
            # Time tag in Unix time (seconds):
            if this_line[0] == "TI":
                tt.append(float(this_line[1]))
                
            # X axis of detector orientation in Galactic coordinates:
            if this_line[0] == "GX":
                this_lonX = np.deg2rad(float(this_line[1])) # radians
                this_latX = np.deg2rad(float(this_line[2])) # radians
                lonX.append(this_lonX)
                latX.append(this_latX)
            
            # Z axis of detector orientation in Galactic coordinates:
            if this_line[0] == "GZ":
                this_lonZ = np.deg2rad(float(this_line[1])) # radians
                this_latZ = np.deg2rad(float(this_line[2])) # radians
                lonZ.append(this_lonZ)
                latZ.append(this_latZ)
            
            # Interaction position information: 
            if (this_line[0] == "CH"):
                
                # First interaction:
                if this_line[1] == "0":
                    v1 = np.array((float(this_line[2]),\
                            float(this_line[3]),float(this_line[4])))
                
                # Second interaction:
                if this_line[1] == "1":
                    v2 = np.array((float(this_line[2]),
                        float(this_line[3]), float(this_line[4])))
                
                    # Compute position vector between first two interactions:
                    dg = v1 - v2
                    dg_x.append(dg[0])
                    dg_y.append(dg[1])
                    dg_z.append(dg[2])

        # Close progress bar:
        pbar.close()
        logger.info("Making COSI data set...")
        logger.info("total events to procecss: " + str(len(erg)))

        # Clear unused memory:
        gc.collect()

        # Initialize arrays:
        logger.info("Initializing arrays...")
        erg = np.array(erg)
        phi = np.array(phi)
        tt = np.array(tt)
        et = np.array(et)
        lonX = np.array(lonX)
        latX = np.array(latX)
        lonZ = np.array(lonZ)
        latZ = np.array(latZ)
        dg_x = np.array(dg_x)
        dg_y = np.array(dg_y)
        dg_z = np.array(dg_z)
 
        # Check if the input data has pointing information, 
        # if not, set dummy values:
        if (use_ori == False) & (len(lonZ)==0):
             logger.warning("WARNING: No pointing information in input data and no ori file.")
             logger.warning("Setting pointing to arbitrary location (Galactic center).")
             lonX = np.array([0]*len(tt))
             latX = np.array([0]*len(tt))
             lonZ = np.array([0]*len(tt))
             latZ = np.array([0]*len(tt))

        # Option to get X and Z pointing information from orientation file:
        if use_ori == True:
            self.instrument_pointing()
            lonX = self.xl_interp(tt)
            latX = self.xb_interp(tt)
            lonZ = self.zl_interp(tt)
            latZ = self.zb_interp(tt)

        # Convert dg vector from 3D cartesian coordinates 
        # to spherical polar coordinates, and then extract distance 
        # b/n first two interactions (in cm), psi (rad), and chi (rad).
        # Note: the resulting angles are latitude/longitude (or elevation/azimuthal).
        conv = astro_co.cartesian_to_spherical(dg_x, dg_y, dg_z)
        dist = conv[0].value 
        psi_loc = conv[1].value 
        chi_loc = conv[2].value

        # Calculate chi_gal and psi_gal from x,y,z coordinates of events:
        xcoords = SkyCoord(lonX*u.rad, latX*u.rad, frame = 'galactic')
        zcoords = SkyCoord(lonZ*u.rad, latZ*u.rad, frame = 'galactic')
        attitude = Attitude.from_axes(x=xcoords, z=zcoords, frame = 'galactic')
        c = SkyCoord(dg_x, dg_y, dg_z, \
            representation_type='cartesian', frame = SpacecraftFrame(attitude = attitude))   
        c_rotated = c.transform_to('galactic')
        chi_gal = np.array(c_rotated.l.deg)
        psi_gal = np.array(c_rotated.b.deg)

        # Change longitudes from 0..360 deg to -180..180 deg
        try:
            lonX[lonX > np.pi] -= 2*np.pi
            lonZ[lonZ > np.pi] -= 2*np.pi
        except:
            logger.warning("Warning: No pointing info to rotate.")

        # Construct Y direction from X and Z direction
        lonlatY = self.construct_scy(np.rad2deg(lonX),np.rad2deg(latX),
                            np.rad2deg(lonZ),np.rad2deg(latZ))
        lonY = np.deg2rad(lonlatY[0])
        latY = np.deg2rad(lonlatY[1])
    
        # Rotate psi_loc to colatitude, measured from positive z direction.
        # This is requred for mhealpy input.
        psi_loc = (np.pi/2.0) - psi_loc 
        
        # Define test values for psi and chi local;
        # this is only for comparing to MEGAlib:
        self.psi_loc_test = psi_loc
        self.chi_loc_test = chi_loc

        # Do the same for psi and chi galactic.
        # First need to convert to radians:
        chi_gal_rad = np.array(c_rotated.l.rad)
        psi_gal_rad = np.array(c_rotated.b.rad)
        
        # Rotate psi_gal_rad to colatitude, measured from positive 
        # z direction:
        psi_gal_rad = (np.pi/2.0) - psi_gal_rad
        self.psi_gal_test = psi_gal_rad
        
        # Rotate chi_gal_test by pi, defined with respect to 
        # the negative x-axis:
        self.chi_gal_test = chi_gal_rad - np.pi
        
        # Make observation dictionary
        logger.info("Making dictionary...")
        cosi_dataset = {'Energies':erg,
                        'TimeTags':tt,
                        'Xpointings (glon,glat)':np.array([lonX,latX]).T,
                        'Ypointings (glon,glat)':np.array([lonY,latY]).T,
                        'Zpointings (glon,glat)':np.array([lonZ,latZ]).T,
                        'Phi':phi,
                        'Chi local':chi_loc,
                        'Psi local':psi_loc,
                        'Distance':dist,
                        'Chi galactic':chi_gal,
                        'Psi galactic':psi_gal} 
        self.cosi_dataset = cosi_dataset

        # Option to write unbinned data to file (either fits or hdf5):
        if output_name != None:
            logger.info("Saving file...")
            self.write_unbinned_output(output_name) 
        
        # Get processing time:
        end_time = time.time()
        processing_time = end_time - start_time
        logger.info("total processing time [s]: " + str(processing_time))
        
        return 

    def instrument_pointing(self):

        """Get pointing information from ori file.
        
        Initializes interpolated functions for lonx, latx, lonz, latz 
        in radians.
        
        Returns
        -------
        xl_interp : scipy:interpolate:interp1d
        xb_interp : scipy:interpolate:interp1d
        zl_interp : scipy:interpolate:interp1d
        zb_interp : scipy:interpolate:interp1d

        Note
        ----
            This method sets the instance attributes, 
            but it does not explicitly return them.
        """

        # Get ori info:
        ori = SpacecraftFile.parse_from_file(self.ori_file)
        time_tags = ori._load_time
        x_pointings = ori.x_pointings
        z_pointings = ori.z_pointings

        # Interpolate:
        self.xl_interp = interpolate.interp1d(time_tags, x_pointings.l.rad, kind='linear')
        self.xb_interp = interpolate.interp1d(time_tags, x_pointings.b.rad, kind='linear')
        self.zl_interp = interpolate.interp1d(time_tags, z_pointings.l.rad, kind='linear')
        self.zb_interp = interpolate.interp1d(time_tags, z_pointings.b.rad, kind='linear')
        
        return 

    def construct_scy(self, scx_l, scx_b, scz_l, scz_b):
    
        """Construct y-coordinate of instrument pointing given x and z directions.
        
        Parameters
        ----------
        scx_l : float
            Longitude of x direction in degrees.
        scx_b : float
            Latitude of x direction in degrees.
        scz_l : float
            Longitude of z direction in degrees.
        scz_b : float
            Latitude of z direction in degrees.

        Returns
        -------
        ra : float
            Right ascension (in degrees) for y-coordinate of instrument pointing.
        dec : float
            Declination (in degrees) for y-coordinate of instrument pointing.

        Note
        ----
            Here, z is the optical axis.
        """
        
        x = self.polar2cart(scx_l, scx_b)
        z = self.polar2cart(scz_l, scz_b)
    
        return self.cart2polar(np.cross(z,x,axis=0))

    def polar2cart(self, ra, dec):
    
        """Coordinate transformation of ra/dec (lon/lat) [phi/theta] 
        polar/spherical coordinates into cartesian coordinates.

        Parameters
        ----------
        ra : float
            Right ascension in degrees. 
        dec: float
            Declination in degrees. 
        
        Returns
        -------
        array
            x, y, and z cartesian coordinates in radians.
        """
        
        x = np.cos(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
        y = np.sin(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
        z = np.sin(np.deg2rad(dec))
    
        return np.array([x,y,z])

    def cart2polar(self, vector):
    
        """Coordinate transformation of cartesian x/y/z values into 
        spherical (deg).

        Parameters
        ----------
        vector : vec
            Vector of x/y/z values.

        Returns
        -------
        ra : float
            Right ascension in degrees.
        dec : float
            Declination in degrees. 
        """
        
        ra = np.arctan2(vector[1],vector[0]) 
        dec = np.arcsin(vector[2])
    
        return np.rad2deg(ra), np.rad2deg(dec)

    def write_unbinned_output(self, output_name):

        """Writes unbinned data file to either fits or hdf5.
        
        Parameters
        ----------
        output_name : str 
            Name of output file. Only include prefix (not file type). 
        """

        # Data units:
        units=['keV','s','rad','rad',
                'rad','rad','rad','rad','cm','deg','deg']
            
        # For fits output: 
        if self.unbinned_output == 'fits':
            table = Table(list(self.cosi_dataset.values()),\
                    names=list(self.cosi_dataset.keys()), \
                    units=units, \
                    meta={'version':cosipy.__version__})
            table.write("%s.fits" %output_name, overwrite=True)
            os.system('gzip -f %s.fits' %output_name)

        # For hdf5 output:
        if self.unbinned_output == 'hdf5':
            with h5py.File('%s.hdf5' %output_name, 'w') as hf:
                for each in list(self.cosi_dataset.keys()):
                    dset = hf.create_dataset(each, data=self.cosi_dataset[each], compression='gzip')        
    
        return

    def get_dict_from_fits(self, input_fits):

        """Constructs dictionary from input fits file.
        
        Parameters
        ----------
        input_fits : str
            Name of input fits file.

        Returns
        -------
        dict
            Dictionary constructed from input fits file.
        """

        # Initialize dictionary:
        this_dict = {}
        
        # Fill dictionary from input fits file:
        hdu = fits.open(input_fits,memmap=True)
        cols = hdu[1].columns
        data = hdu[1].data
        for i in range(0,len(cols)):
            
            this_key = cols[i].name
            this_dict[this_key] = data[this_key]
            
            # Clear unused memory:
            gc.collect()

        return this_dict

    def get_dict_from_hdf5(self, input_hdf5):

        """Constructs dictionary from input hdf5 file
        
        Parameters
        ----------
        input_hdf5 : str
            Name of input hdf5 file. 

        Returns
        -------
        dict
            Dictionary constructed from input hdf5 file.
        """

        # Initialize dictionary:
        this_dict = {}

        # Fill dictionary from input h5fy file:
        hf = h5py.File(input_hdf5,"r")
        keys = list(hf.keys())
        for each in keys:
            this_dict[each] = hf[each][:]

        return this_dict

    def get_dict(self, input_file):

        """Constructs dictionary from input file.
        
        Parameters
        ----------
        input_file : str
            Name of input file. 

        Returns
        -------
        dict
            Dictionary constructed from input file.
        """

        if self.unbinned_output == 'fits':
            this_dict = self.get_dict_from_fits(input_file)
        if self.unbinned_output == 'hdf5':
            this_dict = self.get_dict_from_hdf5(input_file)

        return this_dict

    def select_data_time(self, output_name=None, unbinned_data=None):

        """Applies time cuts to unbinnned data dictionary. 
        
        Parameters
        ----------
        unbinned_data : str, optional
            Name of unbinned dictionary file.
        output_name : str, optional
            Prefix of output file (default is None, in which case no 
            file is saved).
        """
        
        logger.info("Making data selections...")

        # Option to read in unbinned data file:
        if unbinned_data:
            self.cosi_dataset = self.get_dict(unbinned_data)

        # Get time cut index:
        time_array = self.cosi_dataset["TimeTags"]
        time_cut_index = (time_array >= self.tmin) & (time_array < self.tmax)
    
        # Apply cuts to dictionary:
        for key in self.cosi_dataset:

            self.cosi_dataset[key] = self.cosi_dataset[key][time_cut_index]

        # Write unbinned data to file (either fits or hdf5):
        if output_name != None:
            logger.info("Saving file...")
            self.write_unbinned_output(output_name)

        return

    def combine_unbinned_data(self, input_files, output_name=None):

        """Combines input unbinned data files.
        
        Parameters
        ----------
        input_files : list
            List of file names to combine.
        output_name : str, optional
            Prefix of output file.
        """

        self.cosi_dataset = {}
        counter = 0
        for each in input_files:

            logger.info("adding %s..." % each)
    
            # Read dict from hdf5 or fits:
            this_dict = self.get_dict(each)

            # Combine dictionaries:
            if counter == 0:
                for key in this_dict:
                    self.cosi_dataset[key] = this_dict[key]
            
            if counter > 0:
                for key in this_dict:
                    self.cosi_dataset[key] = np.concatenate((self.cosi_dataset[key],this_dict[key]))
                    
            counter =+ 1
            
            # Clear unused memory:
            gc.collect()

        # Write unbinned data to file (either fits or hdf5):
        if output_name != None:
            self.write_unbinned_output(output_name)

        return

    def find_bad_intervals(self, times, values, bad_value=0.0):
        
        """Finds intervals where livetime is 0.0.

        Parameters
        ----------
        times : array
            Array of times from ori file.
        values : array
            Array of livetimes from ori file.        
        bad_value : float or int
            The value that defines a bad time interval. It must match 
            exactly the value in the ori file, including the 
            type (float or int). Default is 0.0. 

        Returns
        -------
        bad_intervals : list of tuples
            List of bad time intervals.
        """
        
        bad_intervals = []
        start = None
    
        for i in range(len(times)-1):
            if values[i] == bad_value:
                if start is None:
                    start = times[i]
            else:
                if start is not None:
                    bad_intervals.append((start.value, times[i].value))
                    start = None
    
        if start is not None:
            bad_intervals.append((start.value, times[-1].value))
    
        return bad_intervals

    def filter_good_data(self, times, bad_intervals):
    
        """Removes entries that fall within bad intervals.

        Parameters
        ----------
        times : array
            Array of photon event times.
        bad_intervals : list
            List of bad intervals defined by livetime = 0.0.
       
        Returns
        -------
        filtered_index : list
            List of indices of good events. 
        """
        
        filtered_index = []

        # Get indices for good times.
        # The inequality below corresponds to left end point in the ori file.
        for i in range(len(times)-1):
            if not any(start < times[i] < end for start, end in bad_intervals):
                filtered_index.append(i)

        return filtered_index

    def cut_SAA_events(self, unbinned_data=None, output_name=None):

        """Cuts events corresponding to SAA passage based on input ori file.

        Parameters
        ----------
        unbinned_data : str, optional
            Name of unbinned dictionary file.
        output_name : str, optional
            Prefix of output file (default is None, in which case no 
            file is saved).
        """

        # Option to read in unbinned data file:
        if unbinned_data:
            self.cosi_dataset = self.get_dict(unbinned_data)

        # Get ori info:
        ori = SpacecraftFile.parse_from_file(self.ori_file)
        
        # Get bad time intervals:
        bti = self.find_bad_intervals(ori._time, ori.livetime)

        # Get indices for good photons:
        time_keep_index = self.filter_good_data(self.cosi_dataset['TimeTags'], bti)

        # Apply cuts to dictionary:
        for key in self.cosi_dataset:

            self.cosi_dataset[key] = self.cosi_dataset[key][time_keep_index]

        # Write unbinned data to file (either fits or hdf5):
        if output_name != None:
            logger.info("Saving file...")
            self.write_unbinned_output(output_name)

        return



import numpy as np
import astropy.units as u
from astropy.units import Quantity
from astropy.coordinates import Galactic
from scipy import integrate

from histpy import Histogram, Axes, Axis, HealpixAxis

from threeML import Band, DiracDelta, Constant, Line, Quadratic, Cubic, Quartic, StepFunction, StepFunctionUpper, Cosine_Prior, Uniform_prior, PhAbs, Gaussian

def get_integrated_spectral_model(spectrum, energy_axis):
    """
    Get the photon fluxes integrated over given energy bins with an input astropy spectral model
        
    Parameters
    ----------
    spectrum : astromodels.functions
        One-dimensional spectral function from astromodels.
    energy_axis : histpy.Axis
        Energy axis defining the energy bins for integration.

    Returns
    -------
    flux : histpy.Histogram
        Histogram of integrated photon fluxes for each energy bin.

    Raises
    ------
    RuntimeError
        If the spectrum is not supported or its units are unknown.

    Notes
    -----
    This function determines the unit of the spectrum, performs the integration
    over each energy bin, and returns the result as a Histogram object.
    """

    from cosipy.threeml import Band_Eflux

    spectrum_unit = None

    for item in spectrum.parameters:
        if getattr(spectrum, item).is_normalization == True:
            spectrum_unit = getattr(spectrum, item).unit
            break
            
    if spectrum_unit == None:
        if isinstance(spectrum, Constant):
            spectrum_unit = spectrum.k.unit
        elif isinstance(spectrum, Line) or isinstance(spectrum, Quadratic) or isinstance(spectrum, Cubic) or isinstance(spectrum, Quartic):
            spectrum_unit = spectrum.a.unit
        elif isinstance(spectrum, StepFunction) or isinstance(spectrum, StepFunctionUpper) or isinstance(spectrum, Cosine_Prior) or isinstance(spectrum, Uniform_prior) or isinstance(spectrum, DiracDelta): 
            spectrum_unit = spectrum.value.unit
        elif isinstance(spectrum, PhAbs):
            spectrum_unit = u.dimensionless_unscaled
        elif isinstance(spectrum, Gaussian):
            spectrum_unit = spectrum.F.unit / spectrum.sigma.unit 
        elif isinstance(spectrum, Band_Eflux):
            spectrum_unit = spectrum.K.unit / spectrum.a.unit
        else:
            try:
                spectrum_unit = spectrum.K.unit
            except:
                raise RuntimeError("Spectrum not yet supported because units of spectrum are unknown.")
                
    if isinstance(spectrum, DiracDelta):
        flux = Quantity([spectrum.value.value * spectrum_unit * lo_lim.unit if spectrum.zero_point.value >= lo_lim/lo_lim.unit and spectrum.zero_point.value <= hi_lim/hi_lim.unit else 0 * spectrum_unit * lo_lim.unit
                         for lo_lim,hi_lim
                         in zip(energy_axis.lower_bounds, energy_axis.upper_bounds)])
    else:
        flux = Quantity([integrate.quad(spectrum, lo_lim/lo_lim.unit, hi_lim/hi_lim.unit)[0] * spectrum_unit * lo_lim.unit
                         for lo_lim,hi_lim
                         in zip(energy_axis.lower_bounds, energy_axis.upper_bounds)])
    
    flux = Histogram(energy_axis, contents = flux)

    return flux

def get_integrated_extended_model(extendedmodel, image_axis, energy_axis):
    """
    Calculate the integrated flux map for an extended source model.

    Parameters
    ----------
    extendedmodel : astromodels.ExtendedSource
        An astromodels extended source model object. This model represents
        the spatial and spectral distribution of an extended astronomical source.
    image_axis : histpy.HealpixAxis
        Spatial axis for the image.
    energy_axis : histpy.Axis
        Energy axis defining the energy bins.

    Returns
    -------
    flux_map : histpy.Histogram
        2D histogram representing the integrated flux map.

    Notes
    -----
    This function first integrates the spectral model over the energy bins,
    then combines it with the spatial distribution to create a 2D flux map.
    """
    
    if not isinstance(image_axis.coordsys, Galactic):
        raise ValueError

    integrated_flux = get_integrated_spectral_model(spectrum = extendedmodel.spectrum.main.shape, energy_axis = energy_axis)

    npix = image_axis.npix
    coords = image_axis.pix2skycoord(np.arange(npix))

    normalized_map = extendedmodel.spatial_shape(coords.l.deg, coords.b.deg) / u.sr

    flux_map = Histogram([image_axis, energy_axis], contents = np.tensordot(normalized_map, integrated_flux.contents, axes = 0))

    return flux_map
    



from .image_deconvolution import ImageDeconvolution

from .image_deconvolution_data_interface_base import ImageDeconvolutionDataInterfaceBase
from .dataIF_COSI_DC2 import DataIF_COSI_DC2

from .model_base import ModelBase
from .allskyimage import AllSkyImageModel

from .deconvolution_algorithm_base import DeconvolutionAlgorithmBase
from .RichardsonLucy import RichardsonLucy
from .RichardsonLucySimple import RichardsonLucySimple

from .exposure_table import SpacecraftAttitudeExposureTable
from .coordsys_conversion_matrix import CoordsysConversionMatrix



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

# Get common version number (https://stackoverflow.com/a/7071358)
import re
VERSIONFILE="cosipy/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(name='cosipy',
      version = verstr,
      author='COSI Team',
      author_email='imc@umd.edu',
      url='https://github.com/cositools/cosipy',
      packages = find_packages(include=["cosipy", "cosipy.*"]),
      install_requires = ["histpy",
                          "mhealpy",
                          "scoords",
                          'astromodels>=2.4.2',
                          'threeml>=2.1.2',
                          'matplotlib<=3.8.3',
                          'numba',
                          'yayc',
                          'awscli',
                          'scipy<=1.13.0'],
      description = "High-level analysis for the COSI telescope data",
      entry_points={"console_scripts":[
          "cosi-response = cosipy.response.FullDetectorResponse:cosi_response",
                              ]},
      
      long_description = long_description,
      long_description_content_type="text/markdown",
      )




from .PointSourceResponse import PointSourceResponse
from .DetectorResponse import DetectorResponse
from .FullDetectorResponse import FullDetectorResponse
from .ExtendedSourceResponse import ExtendedSourceResponse



import os
import copy
import numpy as np
import logging
logger = logging.getLogger(__name__)

from histpy import Histogram

from .deconvolution_algorithm_base import DeconvolutionAlgorithmBase

class RichardsonLucySimple(DeconvolutionAlgorithmBase):
    """
    A class for the original RichardsonLucy algorithm. 
    Basically, this class can be used for testing codes.
    
    An example of parameter is as follows.

    iteration_max: 100
    minimum_flux:
        value: 0.0
        unit: "cm-2 s-1 sr-1"
    background_normalization_optimization: True 
    """

    def __init__(self, initial_model, dataset, mask, parameter):

        DeconvolutionAlgorithmBase.__init__(self, initial_model, dataset, mask, parameter)

        self.do_bkg_norm_optimization = parameter.get('background_normalization_optimization', False)

    def initialization(self):
        """
        initialization before running the image deconvolution
        """
        # clear counter 
        self.iteration_count = 0

        # clear results
        self.results.clear()

        # copy model
        self.model = copy.deepcopy(self.initial_model)

        # calculate exposure map
        self.summed_exposure_map = self.calc_summed_exposure_map()

        # mask setting
        if self.mask is None and np.any(self.summed_exposure_map.contents == 0):
            self.mask = Histogram(self.model.axes, contents = self.summed_exposure_map.contents > 0)
            self.model = self.model.mask_pixels(self.mask)
            logger.info("There are zero-exposure pixels. A mask to ignore them was set.")

        # calculate summed background models for M-step
        if self.do_bkg_norm_optimization:
            self.dict_summed_bkg_model = {}
            for key in self.dict_bkg_norm.keys():
                self.dict_summed_bkg_model[key] = self.calc_summed_bkg_model(key)

    def pre_processing(self):
        """
        pre-processing for each iteration
        """
        pass

    def Estep(self):
        """
        E-step in RL algoritm
        """

        self.expectation_list = self.calc_expectation_list(self.model, dict_bkg_norm = self.dict_bkg_norm)
        logger.debug("The expected count histograms were updated with the new model map.")

    def Mstep(self):
        """
        M-step in RL algorithm.
        """

        ratio_list = [ data.event / expectation for data, expectation in zip(self.dataset, self.expectation_list) ]
        
        # delta model
        sum_T_product = self.calc_summed_T_product(ratio_list)
        self.delta_model = self.model * (sum_T_product/self.summed_exposure_map - 1)
        
        # background normalization optimization
        if self.do_bkg_norm_optimization:
            for key in self.dict_bkg_norm.keys():

                sum_bkg_T_product = self.calc_summed_bkg_model_product(key, ratio_list)
                sum_bkg_model = self.dict_summed_bkg_model[key]

                self.dict_bkg_norm[key] = self.dict_bkg_norm[key] * (sum_bkg_T_product / sum_bkg_model)

    def post_processing(self):
        """
        Post-processing. 
        """
        self.model += self.delta_model
        self.model[:] = np.where(self.model.contents < self.minimum_flux, self.minimum_flux, self.model.contents)

        if self.mask is not None:
            self.model = self.model.mask_pixels(self.mask)

    def register_result(self):
        """
        Register results at the end of each iteration. 
        """
        
        this_result = {"iteration": self.iteration_count, 
                       "model": copy.deepcopy(self.model), 
                       "delta_model": copy.deepcopy(self.delta_model),
                       "background_normalization": copy.deepcopy(self.dict_bkg_norm)}

        # show intermediate results
        logger.info(f'  background_normalization: {this_result["background_normalization"]}')
        
        # register this_result in self.results
        self.results.append(this_result)

    def check_stopping_criteria(self):
        """
        If iteration_count is smaller than iteration_max, the iterative process will continue.

        Returns
        -------
        bool
        """
        if self.iteration_count < self.iteration_max:
            return False
        return True

    def finalization(self):
        """
        finalization after running the image deconvolution
        """
        pass



from ._version import __version__

from .response import DetectorResponse

from .data_io import DataIO
from .data_io import UnBinnedData
from .data_io import BinnedData
from .data_io import ReadTraTest

from .threeml import COSILike
from .threeml import Band_Eflux

from .spacecraftfile import SpacecraftFile

from .ts_map import FastTSMap

from .source_injector import SourceInjector

from .background_estimation import LineBackgroundEstimation
from .background_estimation import ContinuumEstimation
